
const IO = require('../common/io.js');
const IMem = require('./imem');


class ROM extends IMem{
  constructor(pages){
    if (pages < 1)
      throw new RangeError("Memory requires positive page values.");
    super();
    this.__addr = 0;
    this.__map = new Uint8Array(pages * 256);
    // Allow writable events because RAM will inherit ROM and use the same
    // __io object.
    this.__io = new IO([], []);
  }

  get pages(){return this.__map.length / 256;}
  get size(){return this.__map.length;}

  get address(){return this.__addr;}
  set address(a){
    this.__addr = Math.min(this.__map.length - 1, Math.max(0, a));
  }

  get byte(){
    this.__io.triggerRead(this.__addr);
    return this.__map[this.__addr];
  }
  set byte(b){}


  onAddressRead(addr, fn){
    if (addr < 0 || addr >= this.size)
      throw new RangeError("Memory address is out of bounds."); 
    this.__io.onRead(addr, fn);
    return this;
  }


  peek(a){
    a = Math.min(this.__map.length - 1, Math.max(0, a));
    return this.__map[a];
  }


  load(address, data){
    let dc = data;
    if (address < 0 || address >= this.__map.length)
      throw new RangeError("Memory address out of range.");
    if (this.__map.length - address < dc.length)
      dc = dc.slice(0, this.__map.length - address);
    this.__map.set(dc, address);
    return dc.length;
  }


  clearPage(page){
    let addr = (page << 8) & 0xFF00;
    if (addr < 0 || addr >= this.__map.length)
      throw new RangeError("Memory address out of range.");
    this.__map.fill(0, addr, addr + 256);
    return this;
  }


  clear(){
    this.__map.fill(0);
    return this;
  }
}



class RAM extends ROM {
  constructor(pages){
    super(pages);
  }

  get pages(){return this.__map.length/256;}
  get size(){return this.__map.length;}
  get writable(){return true;}


  get address(){return this.__addr;}
  set address(a){
    this.__addr = Math.min(this.__map.length - 1, Math.max(0, a));
  }  

  get byte(){
    this.__io.triggerRead(this.__addr);
    return this.__map[this.__addr];
  }
  set byte(b){
    this.__map[this.__addr] = b & 0xFF;
    this.__io.triggerWrite(this.__addr, b & 0xFF);
  }

  onAddressWrite(addr, fn){
    if (addr < 0 || addr >= this.size)
      throw new RangeError("Memory address is out of bounds.");
    this.__io.onWrite(addr, fn);
    return this;
  }

  poke(a, b){
    a = Math.min(this.__map.length - 1, Math.max(0, a));
    this.__map[a] = b & 0xFF;
    return this;
  }
}


class Shadow extends IMem {
  constructor(pages, rbc){
    if (rbc > (pages * 256))
      throw RangeError("Register Byte Count exceeds available pages defined.");
    super();
    this.__map = new Uint8Array(rbc);
    this.__addr = 0;
    this.__size = pages * 256;

    this.__io = new IO([], []);
  }

  get pages(){return this.__size/256;}
  get size(){return this.__size;}
  get registers(){return this.__map.length;}
  get writable(){return true;}

  get address(){return this.__addr;}
  set address(a){
    if (this.__map)
      this.__addr = Math.min(this.__size - 1, Math.max(0, a));
  }

  get byte(){
    let a = this.__addr % this.__map.length;
    this.__io.triggerRead(this.__addr);
    return this.__map[a];
  }
  set byte(b){
    if (this.__addr >= 0 && this.__addr < this.__map.length){
      this.__map[this.__addr] = b & 0xFF;
      this.__io.triggerWrite(this.__addr, b & 0xFF);
    }
  }

  onAddressRead(addr, fn){
    if (addr < 0 || addr >= this.size)
      throw new RangeError("Memory address is out of bounds."); 
    this.__io.onRead(addr, fn);
    return this;
  }

  onAddressWrite(addr, fn){
    if (addr < 0 || addr >= this.size)
      throw new RangeError("Memory address is out of bounds.");
    let a = this.__addr % this.__map.length;
    this.__io.onWrite(a, fn);
    return this;
  }

  peek(a){
    a = Math.min(this.__size - 1, Math.max(0, a)) % this.__map.length;
    return this.__map[a];
  }

  poke(a, b){
    if (a >= 0 && a < this.__map.length)
      this.__map[a] = b & 0xFF;
    return this;
  }

  load(addr, data){
    let dc = data;
    if (addr < 0 || addr >= this.__size)
      throw new RangeError("Memory address out of range.");
    // We simply quit here because, while it's legal to try to write,
    // there's actually no memory at this point. We're playing pretend :p
    if (addr >= this.__map.length){return 0;}

    if (this.__map.length - addr < dc.length)
      dc = dc.slice(0, this.__map.length - addr);
    this.__map.set(dc, addr);
    return dc.length;
  }


  clearPage(page){
    let addr = (page << 8) & 0xFF00;
    if (addr < 0 || addr >= this.__size)
      throw new RangeError("Memory address out of range.");
    if (addr < this.__map.length){
      let end = (addr + 256 <= this.__map.length) ? 256 : this.__map.length - addr; 
      this.__map.fill(0, addr, end);
    }
    return this;
  }


  clear(){
    if (this.__map)
      this.__map.fill(0);
    return this;
  }
}


module.exports = Object.freeze({
  RAM: RAM,
  ROM: ROM,
  Shadow: Shadow
});





